# Copyright 2021-2022 Zenauth Ltd.
# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Callable, Dict, List, Optional, Set, Union

import httpx
from dataclasses_json import LetterCase, dataclass_json


class Effect(str, Enum):
    DENY = "EFFECT_DENY"
    ALLOW = "EFFECT_ALLOW"


class Source(str, Enum):
    PRINCIPAL = "SOURCE_PRINCIPAL"
    RESOURCE = "SOURCE_RESOURCE"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Principal:
    id: str
    roles: Set[str]
    attr: Dict[str, Any] = field(default_factory=dict)
    policy_version: str = "default"
    scope: str = ""

    def add_attr(self, name: str, value: Any) -> "Principal":
        self.attr[name] = value
        return self


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Resource:
    id: str
    kind: str
    attr: Dict[str, Any] = field(default_factory=dict)
    policy_version: str = "default"
    scope: str = ""

    def add_attr(self, name: str, value: Any) -> "Resource":
        self.attr[name] = value
        return self


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ResourceAction:
    resource: Resource
    actions: Set[str]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ResourceList:
    resources: List[ResourceAction] = field(default_factory=list)

    def add(self, resource: Resource, actions: Set[str]) -> "ResourceList":
        self.resources.append(ResourceAction(resource=resource, actions=actions))
        return self


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class JWT:
    token: str
    key_set_id: Optional[str] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AuxData:
    jwt: JWT


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CheckResourcesRequest:
    request_id: str
    principal: Principal
    resources: ResourceList
    aux_data: Optional[AuxData] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ValidationError:
    path: str
    message: str
    source: Source


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class APIError:
    code: int
    message: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CheckResourcesResult:
    resource: Resource
    actions: Dict[str, Effect]
    validation_errors: Optional[List[ValidationError]] = None

    def is_allowed(self, action: str) -> bool:
        if action in self.actions:
            return self.actions[action] == Effect.ALLOW

        return False


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class CheckResourcesResponse:
    request_id: str
    results: Optional[List[CheckResourcesResult]] = None
    status_code: int = httpx.codes.OK
    status_msg: Optional[APIError] = None

    def failed(self) -> bool:
        return self.status_code != httpx.codes.OK

    def raise_if_failed(self) -> "CheckResourcesResponse":
        if not self.failed():
            return self

        raise CerbosRequestException(self.status_msg)

    def get_resource(
        self, id: str, predicate: Callable[[Resource], bool] = lambda _: True
    ) -> Optional[CheckResourcesResult]:
        if self.failed():
            return None

        return next(
            (r for r in self.results if r.resource.id == id and predicate(r.resource)),
            None,
        )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ResourceDesc:
    kind: str
    attr: Dict[str, Any] = field(default_factory=dict)
    policy_version: str = "default"
    scope: str = ""

    def add_attr(self, name: str, value: Any) -> "ResourceDesc":
        self.attr[name] = value
        return self


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesRequest:
    request_id: str
    action: str
    principal: Principal
    resource: ResourceDesc
    aux_data: Optional[AuxData] = None


Operand = Union[
    "PlanResourcesValue", "PlanResourcesVariable", "PlanResourcesExpression"
]


class PlanResourcesFilterKind(str, Enum):
    ALWAYS_ALLOWED = "KIND_ALWAYS_ALLOWED"
    ALWAYS_DENIED = "KIND_ALWAYS_DENIED"
    CONDITIONAL = "KIND_CONDITIONAL"


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesValue:
    value: Any


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesVariable:
    variable: str


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesExpressionDef:
    operator: str
    operands: List[Operand]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesExpression:
    expression: PlanResourcesExpressionDef


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesFilter:
    kind: PlanResourcesFilterKind
    condition: Optional[Operand] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlanResourcesResponse:
    request_id: str
    action: str
    resource_kind: str
    policy_version: str
    filter: Optional[PlanResourcesFilter] = None
    status_code: int = httpx.codes.OK
    status_msg: Optional[APIError] = None

    def failed(self) -> bool:
        return self.status_code != httpx.codes.OK

    def raise_if_failed(self) -> "PlanResourcesResponse":
        if not self.failed():
            return self

        raise CerbosRequestException(self.status_msg)


class CerbosRequestException(Exception):
    def __init__(self, error: Optional[APIError]):
        msg = "unexpected error" if error is None else error.message
        super(CerbosRequestException, self).__init__(msg)

        self.error = error



# Copyright 2021-2022 Zenauth Ltd.
# SPDX-License-Identifier: Apache-2.0

import logging
import os

import pytest

from cerbos.sdk.client import CerbosClient
from cerbos.sdk.container import CerbosContainer


@pytest.fixture(scope="module", params=["http", "uds"])
def cerbos_client(request, tmp_path_factory):
    policy_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "store")

    container = CerbosContainer(image="ghcr.io/cerbos/cerbos:dev")
    container.with_volume_mapping(policy_dir, "/policies")
    container.with_env("CERBOS_NO_TELEMETRY", "1")

    client = None

    if request.param == "http":
        container.start()
        container.wait_until_ready()

        host = container.http_host()
        logging.info(f"Cerbos Address: {host}")
        client = CerbosClient(host, debug=True)
    elif request.param == "uds":
        sock_dir = tmp_path_factory.mktemp("socket")
        container.with_volume_mapping(sock_dir, "/socket", "rw")
        container.with_command(
            "server --set=server.httpListenAddr=unix:///socket/http.sock --set=server.udsFileMode=0o777"
        )

        container.start()
        container.wait_until_ready()

        client = CerbosClient(f"unix+http://{sock_dir}/http.sock", debug=True)

    yield client

    client.close()
    container.stop()



# Copyright 2021-2022 Zenauth Ltd.
# SPDX-License-Identifier: Apache-2.0

import logging

from cerbos.sdk.client import CerbosClient
from cerbos.sdk.model import *

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    logging.captureWarnings(True)

    with CerbosClient(
        "unix+https:///tmp/cerbos.sock",
        playground_instance="XXY",
        debug=True,
        tls_verify=False,
    ) as c:
        p = Principal(
            "john",
            roles={"employee"},
            policy_version="20210210",
            attr={"department": "marketing", "geography": "GB", "team": "design"},
        )
        r = Resource(
            "XX125",
            "leave_request",
            policy_version="20210210",
            attr={
                "id": "XX125",
                "department": "marketing",
                "geography": "GB",
                "team": "design",
                "owner": "john",
            },
        )
        print(c.is_allowed("view:public", p, r))



# Copyright 2021-2022 Zenauth Ltd.
# SPDX-License-Identifier: Apache-2.0

from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs


class CerbosContainer(DockerContainer):
    HTTP_PORT = 3592
    GRPC_PORT = 3593

    def __init__(self, image: str = "ghcr.io/cerbos/cerbos:latest"):
        super(CerbosContainer, self).__init__(image)
        self.with_exposed_ports(self.HTTP_PORT, self.GRPC_PORT)

    def http_host(self, protocol: str = "http") -> str:
        return f"{protocol}://{self.get_container_host_ip()}:{self.get_exposed_port(self.HTTP_PORT)}"

    def wait_until_ready(self, timeout_secs: float = 30):
        wait_for_logs(self, r"Starting HTTP server at", timeout=timeout_secs)



# Copyright 2021-2022 Zenauth Ltd.
# SPDX-License-Identifier: Apache-2.0

import pytest

from cerbos.sdk.client import CerbosClient, PrincipalContext
from cerbos.sdk.model import *


class TestCerbosClient:
    def test_is_healthy(self, cerbos_client: CerbosClient):
        have = cerbos_client.is_healthy()
        assert have == True

    def test_is_allowed(self, cerbos_client: CerbosClient):
        p = Principal(
            "john",
            roles={"employee"},
            policy_version="20210210",
            attr={"department": "marketing", "geography": "GB", "team": "design"},
        )
        r = Resource(
            "XX125",
            "leave_request",
            policy_version="20210210",
            attr={
                "id": "XX125",
                "department": "marketing",
                "geography": "GB",
                "team": "design",
                "owner": "john",
            },
        )

        have = cerbos_client.is_allowed("view:public", p, r)
        assert have == True

    def test_principal_context_is_allowed(self, principal_ctx: PrincipalContext):
        r = Resource(
            "XX125",
            "leave_request",
            policy_version="20210210",
            attr={
                "id": "XX125",
                "department": "marketing",
                "geography": "GB",
                "team": "design",
                "owner": "john",
            },
        )

        have = principal_ctx.is_allowed("view:public", r)
        assert have == True

    def test_check_resources(self, cerbos_client: CerbosClient):
        p = Principal(
            "john",
            roles={"employee"},
            policy_version="20210210",
            attr={"department": "marketing", "geography": "GB", "team": "design"},
        )

        resources = [
            ResourceAction(
                resource=Resource(
                    "XX125",
                    "leave_request",
                    policy_version="20210210",
                    attr={
                        "id": "XX125",
                        "department": "marketing",
                        "geography": "GB",
                        "team": "design",
                        "owner": "john",
                    },
                ),
                actions={"view:public", "approve"},
            ),
            ResourceAction(
                resource=Resource(
                    "XX225",
                    "leave_request",
                    policy_version="20210210",
                    attr={
                        "id": "XX225",
                        "department": "marketing",
                        "geography": "GB",
                        "team": "design",
                        "owner": "alice",
                        "status": "PENDING_APPROVAL",
                    },
                ),
                actions={"view:private", "approve"},
            ),
        ]

        have = cerbos_client.check_resources(p, ResourceList(resources))
        assert have.failed() == False

        xx125 = have.get_resource(
            "XX125", predicate=lambda r: r.policy_version == "20210210"
        )
        assert xx125 is not None
        assert xx125.is_allowed("view:public") == True
        assert xx125.is_allowed("approve") == False

        xx225 = have.get_resource("XX225")
        assert xx225 is not None
        assert xx225.is_allowed("view:public") == False
        assert xx225.is_allowed("approve") == False

        zz225 = have.get_resource("ZZ225")
        assert zz225 is None

    def test_check_resources_empty_resources(self, cerbos_client: CerbosClient):
        p = Principal(
            "john",
            roles={"employee"},
            policy_version="20210210",
            attr={"department": "marketing", "geography": "GB", "team": "design"},
        )

        have = cerbos_client.check_resources(p, ResourceList())
        assert have.failed() == True
        assert have.status_msg.code == 3

        with pytest.raises(CerbosRequestException):
            have.raise_if_failed()

    def test_plan_resources(self, cerbos_client: CerbosClient):
        p = Principal(
            "maggie",
            roles={"manager"},
            policy_version="20210210",
            attr={
                "department": "marketing",
                "geography": "GB",
                "managed_geographies": "GB",
                "team": "design",
            },
        )

        r = ResourceDesc("leave_request", policy_version="20210210")

        have = cerbos_client.plan_resources("approve", p, r)
        assert have.failed() == False
        assert have.resource_kind == "leave_request"
        assert have.policy_version == "20210210"
        assert have.filter.kind == PlanResourcesFilterKind.CONDITIONAL
        assert have.filter.condition is not None


class TestPrincipalContext:
    def test_is_allowed(self, principal_ctx: PrincipalContext):
        r = Resource(
            "XX125",
            "leave_request",
            policy_version="20210210",
            attr={
                "id": "XX125",
                "department": "marketing",
                "geography": "GB",
                "team": "design",
                "owner": "john",
            },
        )

        have = principal_ctx.is_allowed("view:public", r)
        assert have == True

    def test_check_resources(self, principal_ctx: PrincipalContext):
        resources = [
            ResourceAction(
                resource=Resource(
                    "XX125",
                    "leave_request",
                    policy_version="20210210",
                    attr={
                        "id": "XX125",
                        "department": "marketing",
                        "geography": "GB",
                        "team": "design",
                        "owner": "john",
                    },
                ),
                actions={"view:public", "approve"},
            ),
            ResourceAction(
                resource=Resource(
                    "XX225",
                    "leave_request",
                    policy_version="20210210",
                    attr={
                        "id": "XX225",
                        "department": "marketing",
                        "geography": "GB",
                        "team": "design",
                        "owner": "alice",
                        "status": "PENDING_APPROVAL",
                    },
                ),
                actions={"view:private", "approve"},
            ),
        ]

        have = principal_ctx.check_resources(ResourceList(resources))
        assert have.failed() == False

        xx125 = have.get_resource(
            "XX125", predicate=lambda r: r.policy_version == "20210210"
        )
        assert xx125 is not None
        assert xx125.is_allowed("view:public") == True
        assert xx125.is_allowed("approve") == False

        xx225 = have.get_resource("XX225")
        assert xx225 is not None
        assert xx225.is_allowed("view:public") == False
        assert xx225.is_allowed("approve") == False

        zz225 = have.get_resource("ZZ225")
        assert zz225 is None

    def test_plan_resources(self, principal_ctx: PrincipalContext):
        r = ResourceDesc("leave_request", policy_version="20210210")

        have = principal_ctx.plan_resources("view:public", r)
        assert have.failed() == False
        assert have.resource_kind == "leave_request"
        assert have.policy_version == "20210210"
        assert have.filter.kind == PlanResourcesFilterKind.CONDITIONAL


@pytest.fixture
def principal_ctx(cerbos_client):
    p = Principal(
        "john",
        roles={"employee"},
        policy_version="20210210",
        attr={"department": "marketing", "geography": "GB", "team": "design"},
    )

    return cerbos_client.with_principal(p)
